@@ -1,4 +1,4 @@
-# Copyright 2015-2017 Google Inc. All Rights Reserved.
+# Copyright 2015 Google Inc. All Rights Reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -13,6 +13,8 @@
 # limitations under the License.
 """Computation of split penalties before/between tokens."""
 
+import re
+
 from lib2to3 import pytree
 
 from yapf.yapflib import format_token
@@ -24,10 +26,11 @@ from yapf.yapflib import style
 # TODO(morbo): Document the annotations in a centralized place. E.g., the
 # README file.
 UNBREAKABLE = 1000 * 1000
-NAMED_ASSIGN = 8500
+NAMED_ASSIGN = 11000
 DOTTED_NAME = 4000
 VERY_STRONGLY_CONNECTED = 3500
 STRONGLY_CONNECTED = 3000
+CONNECTED = 500
 
 OR_TEST = 1000
 AND_TEST = 1100
@@ -103,14 +106,23 @@ class _SplitPenaltyAssigner(pytree_visitor.PyTreeVisitor):
     _SetUnbreakable(node.children[colon_idx])
     self.DefaultNodeVisit(node)
     if arrow_idx > 0:
-      _SetSplitPenalty(_LastChildNode(node.children[arrow_idx - 1]), 0)
+      _SetSplitPenalty(
+          pytree_utils.LastLeafNode(node.children[arrow_idx - 1]), 0)
       _SetUnbreakable(node.children[arrow_idx])
       _SetStronglyConnected(node.children[arrow_idx + 1])
 
   def Visit_lambdef(self, node):  # pylint: disable=invalid-name
     # lambdef ::= 'lambda' [varargslist] ':' test
     # Loop over the lambda up to and including the colon.
-    if style.Get('ALLOW_MULTILINE_LAMBDAS'):
+    allow_multiline_lambdas = style.Get('ALLOW_MULTILINE_LAMBDAS')
+    if not allow_multiline_lambdas:
+      for child in node.children:
+        if pytree_utils.NodeName(child) == 'COMMENT':
+          if re.search(r'pylint:.*disable=.*\bg-long-lambda', child.value):
+            allow_multiline_lambdas = True
+            break
+
+    if allow_multiline_lambdas:
       _SetStronglyConnected(node)
     else:
       self._SetUnbreakableOnChildren(node)
@@ -133,18 +145,34 @@ class _SplitPenaltyAssigner(pytree_visitor.PyTreeVisitor):
       if isinstance(child, pytree.Leaf) and child.value == ',':
         _SetUnbreakable(child)
       index += 1
+    for child in node.children:
+      if pytree_utils.NodeName(child) == 'atom':
+        _IncreasePenalty(child, CONNECTED)
 
   def Visit_argument(self, node):  # pylint: disable=invalid-name
     # argument ::= test [comp_for] | test '=' test  # Really [keyword '='] test
     self.DefaultNodeVisit(node)
-
     index = 1
     while index < len(node.children) - 1:
       child = node.children[index]
       if isinstance(child, pytree.Leaf) and child.value == '=':
-        _SetSplitPenalty(_FirstChildNode(node.children[index]), NAMED_ASSIGN)
         _SetSplitPenalty(
-            _FirstChildNode(node.children[index + 1]), NAMED_ASSIGN)
+            pytree_utils.FirstLeafNode(node.children[index]), NAMED_ASSIGN)
+        _SetSplitPenalty(
+            pytree_utils.FirstLeafNode(node.children[index + 1]), NAMED_ASSIGN)
+      index += 1
+
+  def Visit_tname(self, node):  # pylint: disable=invalid-name
+    # tname ::= NAME [':' test]
+    self.DefaultNodeVisit(node)
+    index = 1
+    while index < len(node.children) - 1:
+      child = node.children[index]
+      if isinstance(child, pytree.Leaf) and child.value == ':':
+        _SetSplitPenalty(
+            pytree_utils.FirstLeafNode(node.children[index]), NAMED_ASSIGN)
+        _SetSplitPenalty(
+            pytree_utils.FirstLeafNode(node.children[index + 1]), NAMED_ASSIGN)
       index += 1
 
   def Visit_dotted_name(self, node):  # pylint: disable=invalid-name
@@ -164,7 +192,6 @@ class _SplitPenaltyAssigner(pytree_visitor.PyTreeVisitor):
 
   def Visit_trailer(self, node):  # pylint: disable=invalid-name
     # trailer ::= '(' [arglist] ')' | '[' subscriptlist ']' | '.' NAME
-    self.DefaultNodeVisit(node)
     if node.children[0].value == '.':
       self._SetUnbreakableOnChildren(node)
       _SetSplitPenalty(node.children[1], DOTTED_NAME)
@@ -173,12 +200,17 @@ class _SplitPenaltyAssigner(pytree_visitor.PyTreeVisitor):
       _SetSplitPenalty(node.children[1], VERY_STRONGLY_CONNECTED)
     elif len(node.children) == 3:
       name = pytree_utils.NodeName(node.children[1])
-      if name == 'power':
-        if pytree_utils.NodeName(node.children[1].children[0]) != 'atom':
-          # Don't split an argument list with one element if at all possible.
-          _SetStronglyConnected(node.children[1], node.children[2])
+      if name in {'argument', 'comparison'}:
+        # Don't split an argument list with one element if at all possible.
+        _SetStronglyConnected(node.children[1])
+        if (len(node.children[1].children) > 1 and
+            pytree_utils.NodeName(node.children[1].children[1]) == 'comp_for'):
+          # Don't penalize splitting before a comp_for expression.
+          _SetSplitPenalty(pytree_utils.FirstLeafNode(node.children[1]), 0)
+        else:
           _SetSplitPenalty(
-              _FirstChildNode(node.children[1]), ONE_ELEMENT_ARGUMENT)
+              pytree_utils.FirstLeafNode(node.children[1]),
+              ONE_ELEMENT_ARGUMENT)
       elif (pytree_utils.NodeName(node.children[0]) == 'LSQB' and
             len(node.children[1].children) > 2 and
             (name.endswith('_test') or name.endswith('_expr'))):
@@ -191,19 +223,26 @@ class _SplitPenaltyAssigner(pytree_visitor.PyTreeVisitor):
                         (name.endswith('_expr') and
                          style.Get('SPLIT_BEFORE_BITWISE_OPERATOR')))
         if split_before:
-          _SetSplitPenalty(_LastChildNode(node.children[1].children[1]), 0)
+          _SetSplitPenalty(
+              pytree_utils.LastLeafNode(node.children[1].children[1]), 0)
         else:
-          _SetSplitPenalty(_FirstChildNode(node.children[1].children[2]), 0)
+          _SetSplitPenalty(
+              pytree_utils.FirstLeafNode(node.children[1].children[2]), 0)
 
         # Don't split the ending bracket of a subscript list.
         _SetVeryStronglyConnected(node.children[-1])
       elif name not in {
           'arglist', 'argument', 'term', 'or_test', 'and_test', 'comparison',
-          'atom'
+          'atom', 'power'
       }:
         # Don't split an argument list with one element if at all possible.
         _SetStronglyConnected(node.children[1], node.children[2])
 
+      if name == 'arglist':
+        _SetStronglyConnected(node.children[-1])
+
+    self.DefaultNodeVisit(node)
+
   def Visit_power(self, node):  # pylint: disable=invalid-name,missing-docstring
     # power ::= atom trailer* ['**' factor]
     self.DefaultNodeVisit(node)
@@ -252,13 +291,15 @@ class _SplitPenaltyAssigner(pytree_visitor.PyTreeVisitor):
           subtypes = pytree_utils.GetNodeAnnotation(
               trailer.children[0], pytree_utils.Annotation.SUBTYPE)
           if subtypes and format_token.Subtype.SUBSCRIPT_BRACKET in subtypes:
-            _SetStronglyConnected(_FirstChildNode(trailer.children[1]))
+            _SetStronglyConnected(
+                pytree_utils.FirstLeafNode(trailer.children[1]))
 
-          last_child_node = _LastChildNode(trailer)
+          last_child_node = pytree_utils.LastLeafNode(trailer)
           if last_child_node.value.strip().startswith('#'):
             last_child_node = last_child_node.prev_sibling
           if not style.Get('DEDENT_CLOSING_BRACKETS'):
-            if _LastChildNode(last_child_node.prev_sibling).value != ',':
+            last = pytree_utils.LastLeafNode(last_child_node.prev_sibling)
+            if last.value != ',':
               if last_child_node.value == ']':
                 _SetUnbreakable(last_child_node)
               else:
@@ -280,7 +321,7 @@ class _SplitPenaltyAssigner(pytree_visitor.PyTreeVisitor):
 
   def Visit_comp_for(self, node):  # pylint: disable=invalid-name
     # comp_for ::= 'for' exprlist 'in' testlist_safe [comp_iter]
-    _SetSplitPenalty(_FirstChildNode(node), 0)
+    _SetSplitPenalty(pytree_utils.FirstLeafNode(node), 0)
     _SetStronglyConnected(*node.children[1:])
     self.DefaultNodeVisit(node)
 
@@ -298,10 +339,11 @@ class _SplitPenaltyAssigner(pytree_visitor.PyTreeVisitor):
     index = 1
     while index + 1 < len(node.children):
       if style.Get('SPLIT_BEFORE_LOGICAL_OPERATOR'):
-        _DecrementSplitPenalty(_FirstChildNode(node.children[index]), OR_TEST)
+        _DecrementSplitPenalty(
+            pytree_utils.FirstLeafNode(node.children[index]), OR_TEST)
       else:
         _DecrementSplitPenalty(
-            _FirstChildNode(node.children[index + 1]), OR_TEST)
+            pytree_utils.FirstLeafNode(node.children[index + 1]), OR_TEST)
       index += 2
 
   def Visit_and_test(self, node):  # pylint: disable=invalid-name
@@ -311,10 +353,11 @@ class _SplitPenaltyAssigner(pytree_visitor.PyTreeVisitor):
     index = 1
     while index + 1 < len(node.children):
       if style.Get('SPLIT_BEFORE_LOGICAL_OPERATOR'):
-        _DecrementSplitPenalty(_FirstChildNode(node.children[index]), AND_TEST)
+        _DecrementSplitPenalty(
+            pytree_utils.FirstLeafNode(node.children[index]), AND_TEST)
       else:
         _DecrementSplitPenalty(
-            _FirstChildNode(node.children[index + 1]), AND_TEST)
+            pytree_utils.FirstLeafNode(node.children[index + 1]), AND_TEST)
       index += 2
 
   def Visit_not_test(self, node):  # pylint: disable=invalid-name
@@ -326,8 +369,10 @@ class _SplitPenaltyAssigner(pytree_visitor.PyTreeVisitor):
     # comparison ::= expr (comp_op expr)*
     self.DefaultNodeVisit(node)
     if len(node.children) == 3 and _StronglyConnectedCompOp(node):
-      _SetSplitPenalty(_FirstChildNode(node.children[1]), STRONGLY_CONNECTED)
-      _SetSplitPenalty(_FirstChildNode(node.children[2]), STRONGLY_CONNECTED)
+      _SetSplitPenalty(
+          pytree_utils.FirstLeafNode(node.children[1]), STRONGLY_CONNECTED)
+      _SetSplitPenalty(
+          pytree_utils.FirstLeafNode(node.children[2]), STRONGLY_CONNECTED)
     else:
       _IncreasePenalty(node, COMPARISON)
 
@@ -348,7 +393,7 @@ class _SplitPenaltyAssigner(pytree_visitor.PyTreeVisitor):
           _SetSplitPenalty(child, style.Get('SPLIT_PENALTY_BITWISE_OPERATOR'))
         else:
           _SetSplitPenalty(
-              _FirstChildNode(node.children[index + 1]),
+              pytree_utils.FirstLeafNode(node.children[index + 1]),
               style.Get('SPLIT_PENALTY_BITWISE_OPERATOR'))
       index += 1
 
@@ -367,15 +412,35 @@ class _SplitPenaltyAssigner(pytree_visitor.PyTreeVisitor):
     self.DefaultNodeVisit(node)
     _IncreasePenalty(node, SHIFT_EXPR)
 
+  _ARITH_OPS = frozenset({'PLUS', 'MINUS'})
+
   def Visit_arith_expr(self, node):  # pylint: disable=invalid-name
     # arith_expr ::= term (('+'|'-') term)*
     self.DefaultNodeVisit(node)
     _IncreasePenalty(node, ARITH_EXPR)
 
+    index = 1
+    while index < len(node.children) - 1:
+      child = node.children[index]
+      if pytree_utils.NodeName(child) in self._ARITH_OPS:
+        next_node = pytree_utils.FirstLeafNode(node.children[index + 1])
+        _SetSplitPenalty(next_node, ARITH_EXPR)
+      index += 1
+
+  _TERM_OPS = frozenset({'STAR', 'AT', 'SLASH', 'PERCENT', 'DOUBLESLASH'})
+
   def Visit_term(self, node):  # pylint: disable=invalid-name
     # term ::= factor (('*'|'@'|'/'|'%'|'//') factor)*
-    _IncreasePenalty(node, TERM)
     self.DefaultNodeVisit(node)
+    _IncreasePenalty(node, TERM)
+
+    index = 1
+    while index < len(node.children) - 1:
+      child = node.children[index]
+      if pytree_utils.NodeName(child) in self._TERM_OPS:
+        next_node = pytree_utils.FirstLeafNode(node.children[index + 1])
+        _SetSplitPenalty(next_node, TERM)
+      index += 1
 
   def Visit_factor(self, node):  # pyline: disable=invalid-name
     # factor ::= ('+'|'-'|'~') factor | power
@@ -390,10 +455,10 @@ class _SplitPenaltyAssigner(pytree_visitor.PyTreeVisitor):
     if node.children[0].value == '(':
       if node.children[-1].value == ')':
         if pytree_utils.NodeName(node.parent) == 'if_stmt':
-          _SetSplitPenalty(node.children[-1], UNBREAKABLE)
+          _SetSplitPenalty(node.children[-1], STRONGLY_CONNECTED)
         else:
           if len(node.children) > 2:
-            _SetSplitPenalty(_FirstChildNode(node.children[1]), EXPR)
+            _SetSplitPenalty(pytree_utils.FirstLeafNode(node.children[1]), EXPR)
           _SetSplitPenalty(node.children[-1], ATOM)
     elif node.children[0].value in '[{' and len(node.children) == 2:
       # Keep empty containers together if we can.
@@ -408,7 +473,7 @@ class _SplitPenaltyAssigner(pytree_visitor.PyTreeVisitor):
         prev_was_comma = True
       else:
         if prev_was_comma:
-          _SetSplitPenalty(_FirstChildNode(child), 0)
+          _SetSplitPenalty(pytree_utils.FirstLeafNode(child), 0)
         prev_was_comma = False
 
   ############################################################################
@@ -460,7 +525,7 @@ def _SetExpressionPenalty(node, penalty):
       for child in node.children:
         RecExpression(child, first_child_leaf)
 
-  RecExpression(node, _FirstChildNode(node))
+  RecExpression(node, pytree_utils.FirstLeafNode(node))
 
 
 def _IncreasePenalty(node, amt):
@@ -480,7 +545,7 @@ def _IncreasePenalty(node, amt):
       for child in node.children:
         RecExpression(child, first_child_leaf)
 
-  RecExpression(node, _FirstChildNode(node))
+  RecExpression(node, pytree_utils.FirstLeafNode(node))
 
 
 def _RecAnnotate(tree, annotate_name, annotate_value):
@@ -506,8 +571,8 @@ def _RecAnnotate(tree, annotate_name, annotate_value):
 def _StronglyConnectedCompOp(op):
   if (len(op.children[1].children) == 2 and
       pytree_utils.NodeName(op.children[1]) == 'comp_op' and
-      _FirstChildNode(op.children[1]).value == 'not' and
-      _LastChildNode(op.children[1]).value == 'in'):
+      pytree_utils.FirstLeafNode(op.children[1]).value == 'not' and
+      pytree_utils.LastLeafNode(op.children[1]).value == 'in'):
     return True
   if (isinstance(op.children[1], pytree.Leaf) and
       op.children[1].value in {'==', 'in'}):
@@ -545,15 +610,3 @@ def _AllowBuilderStyleCalls(node):
       if prev_child.lineno != child.lineno:
         _SetSplitPenalty(child, 0)
     prev_child = child
-
-
-def _FirstChildNode(node):
-  if isinstance(node, pytree.Leaf):
-    return node
-  return _FirstChildNode(node.children[0])
-
-
-def _LastChildNode(node):
-  if isinstance(node, pytree.Leaf):
-    return node
-  return _LastChildNode(node.children[-1])
